import os
import time
import copy
import numpy as np
import torch
import torch.nn as nn
import torch.nn.functional as F


def loss_cos(args, model, x_adv, x_ori, adv_label, distance):
    """
    * loss function
    :param args: parameters
    :param model: target model
    :param x_adv: adversarial sample at the boundary
    :param x_ori: target sample
    :param adv_label: adversarial label yk
    :param distance: distance of last step
    :return cos: [-1, 0]
    """
    # vector of original data and adversarial sample
    x_sub = x_adv - x_ori
    x_sub = x_sub.view(-1)
    x_sub = x_sub / torch.norm(x_sub, p=2)

    # the normal vector of decision boundary on x_adv
    grad_xadv = gradient_compute_boundary(args, model, x_adv, adv_label, distance)
    grad_xadv = grad_xadv.view(-1)
    grad_xadv = grad_xadv / torch.norm(grad_xadv, p=2)

    # compute cosine similarity
    cos = -torch.cosine_similarity(x_sub, grad_xadv, dim=-1)

    return cos


def gradient_compute_boundary(args, model, sample, adv_label, distance):
    """
    * estimate the normal vector of decision boundary with MCMC
    :param args: parameters
    :param model: target model
    :param sample: adversarial sample at the boundary
    :param adv_label: adversarial label yk
    :param distance: distance of last step
    :return gradf: normal vector of decision boundary
    """
    sample = sample.squeeze(0)
    # set number of disturbed groups B
    num_evals = args.num_evals_boundary
    # set step size $\delta$
    delta = 1 / len(sample.shape) * distance
    if delta > 0.2:
        delta = 0.2

    # get the noise
    noise_shape = [num_evals] + list(sample.shape)
    rv = torch.randn(*noise_shape, device=args.device)
    rv = rv / torch.sqrt(torch.sum(rv ** 2, dim=(1, 2, 3), keepdim=True))

    # get the noise sample
    perturbed = sample + delta * rv

    # model prediction
    # compute $\Phi(\bar{x}+\delta ub)$
    decisions = decision_function(args, model, perturbed, adv_label)
    decision_shape = [len(decisions)] + [1] * len(sample.shape)
    fval = 2 * decisions.reshape(decision_shape) - 1.0

    # compute gradf as the normal vector N
    if torch.mean(fval) == torch.tensor(1.0):
        gradf = torch.mean(rv, dim=0)
    elif torch.mean(fval) == torch.tensor(-1.0):
        gradf = - torch.mean(rv, dim=0)
    else:
        fval -= torch.mean(fval)
        gradf = torch.mean(fval * rv, dim=0)

    return gradf


def binary_search(args, model, x_0, x_random, adv_label, tol=1e-5):
    """
    * binary search function to project adversarial sample onto the decision boundary
    :param args: parameters
    :param model: target model
    :param x_0: target sample
    :param x_random: adversarial sample $\tilde{x}$
    :param adv_label: adversarial label yk
    :param tol: the threshold value
    :return adv: adversarial sample $\bar{x}$ at the boundary
    """
    if args.data_name == 'TinyImageNet': tol = 1e-4

    adv = x_random
    cln = x_0

    while True:

        mid = (cln + adv) / 2.0

        if decision_function(args, model, mid, adv_label):
            adv = mid
        else:
            cln = mid

        if torch.norm(adv - cln).cpu().numpy() < tol:
            break

    return adv


def decision_function(args, model, images, adv_label, batch_size=100):
    """
    Decision function output 1 on the desired side of the boundary, 0 otherwise
	:param args: parameters
    :param model: target model
    :param images: noise sample
    :param adv_label: adversarial label yk
    :param batch_size: the size of a batch for reducing time of prediction
    :return: whether the data point is adversarial (1:yes 0:no)
    """
    model.train(mode=False)

    # the prediction is made directly if there is only one data,
    if len(images) == 1:
        with torch.no_grad():
            predict_label = torch.argmax(model(images), dim=1).reshape(len(images), 1)

    else:
        results_list = []

        # Run prediction with batch processing
        num_batch = int(np.ceil(len(images) / float(batch_size)))
        for m in range(num_batch):
            # Batch indexes
            begin, end = (
                m * batch_size,
                min((m + 1) * batch_size, images.shape[0]),
            )
            with torch.no_grad():
                output = model(images[begin:end])
            output = output.detach().cpu().numpy().astype(np.float32)
            results_list.append(output)

        results = np.vstack(results_list)
        predict = torch.from_numpy(results).to(args.device)
        predict_label = torch.argmax(predict, dim=1).reshape(len(images), 1)

    target_label = torch.zeros((len(images), 1), device=args.device)
    for i in range((len(images))):
        target_label[i, 0] = adv_label

    return predict_label == target_label


def adv_attack(args, model, x_sample, y_sample, tgt_sample, adv_label):
    """
    MinAD function
    * 1.get the sample $\bar{xi}$ at the boundary by binary search;
    * 2.estimate normal vector of the boundary at $\bar{xi}$;
    * 3.update the sample $\tilde{xi+1}$ by SGD;
    * 4.get the sample $\bar{xi+1}$ at the boundary by binary search again.
    :param args: parameters
    :param model: target model
    :param x_sample: target sample x
    :param y_sample: label of x
    :param tgt_sample: initial adversarial sample $\tilde{x0}$
    :param adv_label: adversarial label yk
    :return adv_sample: the sample $\bar{xI}$ at the boundary
    :return distance: the distance between target sample x and model boundary
    """
    # get $\bar{x0}$ at the boundary
    adv_init = binary_search(args, model, x_sample, tgt_sample, adv_label)
    adv_update = adv_init
    ori_sample = x_sample

    distance_init = torch.norm(adv_init - ori_sample, p=2)
    distance_value = distance_init

    # main loop of update
    lr1 = args.attack_lr_begin
    temp_list = []
    for attack_epoch in range(args.num_attack_iter):

        adv_sample = copy.deepcopy(adv_update)
        distance = copy.deepcopy(distance_value)

        # estimate normal vector
        adv_sample.requires_grad = True
        loss = loss_cos(args, model, adv_sample, ori_sample, adv_label, distance)
        # logger.info("MinAD: epoch:{}, adv_label:{}, loss:{}, distance:{}".format(attack_epoch, torch.argmax(model(adv_sample)), loss, distance))
        loss.backward()
        grads = adv_sample.grad
        grads = grads / torch.norm(grads, p=2)
        adv_sample.requires_grad = False

        # update the sample
        # lr1: learning rate for update
        # lr2: we set a new learning rate to ensure that the updated data is adversarial
        temp = 0
        adv_sample -= lr1 * grads

        # set the auxiliary debugging function: if the updated data is not adversarial, the update step may be too large
        lr2 = lr1
        while not decision_function(args, model, adv_sample, adv_label):
            temp = 1
            lr2 = lr2 / 2
            adv_sample += lr2 * grads
            if lr2 < args.attack_lr_end / 2:
                break
        temp_list.append(temp)

        # decay of learning rate
        if not temp and lr1 > args.attack_lr_end \
                or sum(temp_list[attack_epoch-4: attack_epoch]) == 4:
            lr1 /= 2
            # set threshold
            if lr1 < args.attack_lr_end:
                lr1 = args.attack_lr_end

        # get the sample at the boundary by binary search
        adv_sample = binary_search(args, model, ori_sample, adv_sample, adv_label)

        # compute the distance between target sample x and final adversarial sample at the boundary
        distance_value = torch.norm(adv_sample - ori_sample, p=2)

        adv_update = copy.deepcopy(adv_sample)

    # logger.info("MinAD: ori_sample:{}, adv_label:{}, distance:{}".format(y_sample, adv_label, distance_value))
    d_linf = torch.norm(adv_sample - ori_sample, p=np.inf)
    d_l2 = distance_value
    d_l1 = torch.norm(adv_sample - ori_sample, p=1)

    return adv_sample, torch.tensor([d_linf, d_l2, d_l1], device=args.device)


def set_target(data_name, model, loader, num_class, device):
    ## Find the corresponding adversarial target sample($\tilde{x}_0$) for each class

    tgt_list = torch.zeros(num_class, 1, 3, 32, 32, device=device) \
        if data_name == 'CIFAR10' else torch.zeros(num_class, 1, 3, 224, 224, device=device)

    for tgt_label in range(num_class):
        tgt_label = torch.tensor(tgt_label)
        tgt_temp = 0
        for idb, (XX_tgt, YY_tgt) in enumerate(loader):
            if tgt_temp:
                break
            for i_yy in range(len(YY_tgt)):
                # find the data
                if YY_tgt[i_yy] == tgt_label:
                    tgt_temp = 1
                    tgt_tgt = XX_tgt[i_yy].unsqueeze(0).to(device)
                    # add the data
                    tgt_list[int(tgt_label.item())] = tgt_tgt
                    break

    assert tgt_list.shape[0] == num_class

    return tgt_list


def get_black_vulnerabilty(args, model, inputs, tgt_list, adv_class):

    # number of examples for adversarial example generation
    args.num_samples = inputs.shape[0]

    # list for saving distance
    distance_list = torch.zeros(args.num_samples, 3, device=args.device)
    # list for saving adversarial examples
    adv_list = torch.zeros(args.num_samples, 3, 32, 32, device=args.device) \
        if args.data_name == 'CIFAR10' else torch.zeros(args.num_samples, 3, 224, 224, device=args.device)

    for xx_idx in range(args.num_samples):
        start = time.time()

        ## inputs setting
        # attack each data in inputs
        # cifar10: X_ [args.num_samples, 3, 32, 32]->[1, 3, 32, 32]
        # cifar10: model(X)_ [1, 10]->[10, ]
        xx_tgt = inputs[xx_idx].unsqueeze(0).to(args.device)
        yy_tgt = torch.argmax(model(xx_tgt))

        adv_label = adv_class[xx_idx].unsqueeze(0).to(args.device)

        tgt_tgt = tgt_list[adv_label.item()]
        # logger.info('MinAD attack process|| data_num:{}, yy_label:{}, adv_label:{}, distance:{}'.
        #           format(xx_idx, yy_tgt, adv_label, torch.norm(tgt_tgt - xx_tgt, p=2)))

        adv, d = adv_attack(args, model, xx_tgt, yy_tgt, tgt_tgt, adv_label)
        distance_list[xx_idx] = d
        adv_list[xx_idx] = adv.squeeze(0)

        print('MinAD attack|| data_num:{}, distance:{}, t:{}'.format(xx_idx, distance_list[xx_idx], time.time() - start))

    full_adv = adv_list

    return full_adv

